package edu.fsu.cs.hannigan.graphingwidget;

import java.util.List;

import edu.fsu.cs.hannigan.graphingwidget.GraphingThread;

import android.content.Context;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.Paint.Cap;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.util.AttributeSet;
import android.view.SurfaceHolder;
import android.view.SurfaceView;


public class GraphingWidget extends SurfaceView implements SurfaceHolder.Callback, SensorEventListener{
	public static final int DEFAULT_HEIGHT = 100;
	public static final int DEFAULT_WIDTH = 150;
	public static final int BEAT_GAP = 5;
	
	// graphing modes
	public static final int ACCELEROMETER = 0;
	public static final int HEART_BEAT = 1;
	public static final int CUSTOM = 2;
	private int mode = 0;

	
	// Accelerometer stuff
	private SensorManager mySManager;
	private Sensor accSensor;
	private List<Sensor> sensors;

	// variables for coordinates of graph
	private float _x = 0, prevX = 0;
	private float _y = 0, prevY = 0, base_y = 0;
	private float _yofX = 0, _yofY = 0,	_yofZ = 0;
	private float prevYofX = 0,	prevYofY = 0, prevYofZ = 0;

	// custom values for graphing variability
	private float speedofX = 1;
	private int intensityofY = 5;

	// Drawing stuff
	private GraphingThread _thread;
	private Paint paint = new Paint();
	private boolean cleanCanvas = true;
	
	public GraphingWidget(Context context) {
		super(context);
		initGraphingWidget();
	}

	public GraphingWidget(Context context, AttributeSet attrs){
		super(context, attrs);
		initGraphingWidget();
	}

	private final void initGraphingWidget() {
		getHolder().addCallback(this);
		_thread = new GraphingThread(getHolder(), this);
		
		// set and register accelerometer listener!
		mySManager = (SensorManager) getContext().getSystemService(Context.SENSOR_SERVICE);
		sensors = mySManager.getSensorList(Sensor.TYPE_ACCELEROMETER);

		if(sensors.size() > 0)
		{	accSensor = sensors.get(0); }
		mySManager.registerListener(this, accSensor, 2);
	}

	@Override
	protected void onMeasure(int widthMeasureSpec, int heightMeasureSpec){
		setMeasuredDimension(measureWidth(widthMeasureSpec),
				measureHeight(heightMeasureSpec));
	}

	public int measureWidth(int measureSpec){
		int result = 0;
		int specMode = MeasureSpec.getMode(measureSpec);
		int specSize = MeasureSpec.getSize(measureSpec);

		if (specMode == MeasureSpec.EXACTLY) {
			// We were told how big to be
			result = specSize;
		} else {
			result = DEFAULT_WIDTH;
		}

		return result;
	}

	public int measureHeight(int measureSpec){
		int result = 0;
		int specMode = MeasureSpec.getMode(measureSpec);
		int specSize = MeasureSpec.getSize(measureSpec);

		if (specMode == MeasureSpec.EXACTLY) {
			// We were told how big to be
			result = specSize;
		} else {
			result = DEFAULT_HEIGHT;
		}

		return result;
	}

	
	@Override
	protected void onDraw(Canvas canvas) {
		super.onDraw(canvas);

		// This makes sure that both buffers have been cleared.
		if(cleanCanvas)
		{     
			//canvas.drawColor(Color.WHITE);
			canvas.drawColor(Color.BLACK);
			cleanCanvas = false;
		}

		// Reset the graph after it reaches the end of the window.
		if(_x > this.getWidth())
		{
			_x = 0; 
			prevX = 0;
			cleanCanvas = true;
			//canvas.drawColor(Color.WHITE);
			canvas.drawColor(Color.BLACK);
		}
		
		switch(mode)
		{
		case ACCELEROMETER:
			paint.setStrokeCap(Cap.ROUND);
			paint.setColor(Color.RED);
			canvas.drawLine(prevX, 100 - prevYofX, _x, 100 - _yofX, paint);
			paint.setColor(Color.BLUE);
			canvas.drawLine(prevX, 100 - prevYofY, _x, 100 - _yofY, paint);
			paint.setColor(Color.GREEN);
			canvas.drawLine(prevX, 100 - prevYofZ, _x, 100 - _yofZ , paint);

			prevYofX = _yofX;
			prevYofY = _yofY;
			prevYofZ = _yofZ;
			prevX = _x;
			_x += speedofX;
			break;
		case HEART_BEAT:
			base_y = this.getHeight()/2;
			paint.setColor(Color.RED);
			paint.setStrokeCap(Cap.ROUND);
			if(_y != base_y)  // then a beat has been received 
			{
				_x += speedofX+BEAT_GAP;
				canvas.drawLine(prevX, base_y, _x, base_y-_y, paint);
				prevX = _x;
				_x += speedofX+BEAT_GAP;
				prevY = base_y-_y;
				canvas.drawLine(prevX, prevY, _x, base_y+(_y/2), paint);
				prevX = _x;
				prevY = base_y+(_y/2);
				_x += speedofX+BEAT_GAP;
				canvas.drawLine(prevX, prevY, _x, base_y, paint);
				prevX = _x;
				prevY = _y;
				_x += speedofX+BEAT_GAP;
				_y = base_y;
				canvas.drawLine(prevX, base_y, _x, base_y, paint);
			}
			else
				canvas.drawLine(prevX, base_y, _x, base_y, paint);
				prevX = _x;
				_x += speedofX;
			break;
		case CUSTOM:
			paint.setColor(Color.RED);
			canvas.drawLine(prevX, canvas.getHeight() - prevY, _x, canvas.getHeight() - _y, paint);
			//canvas.drawLine(prevX, 100 - prevY, _x, 100 - _y, paint);

			prevX = _x;
			prevY = _y;
			_x += speedofX;
			break;			
		}

	}

	public void surfaceChanged(SurfaceHolder holder, int format, int width,
			int height) {	}

	public void surfaceCreated(SurfaceHolder holder) {
		_thread = new GraphingThread(getHolder(), this);
		_thread.setRunning(true);
		_thread.start();
	}

	public void surfaceDestroyed(SurfaceHolder holder) {
		boolean retry = true;
		_thread.setRunning(false);
		mySManager.unregisterListener(this);
		while (retry) {
			try {
				_thread.join();
				retry = false;
			} catch (InterruptedException e) {
				// we will try it again and again...
			}
		}
	}

	public void onAccuracyChanged(Sensor sensor, int accuracy) {	
	}

	public void onSensorChanged(SensorEvent event) {
		_yofX = event.values[0]*intensityofY;
		_yofY = event.values[1]*intensityofY;
		_yofZ = event.values[2]*intensityofY;
	}	
	
	// GET AND SET FUNCTIONS!!
	public void setMode(int val)
	{	mode = val; }
	public void setY(float val)
	{	_y = val*intensityofY; }
	public void setSpeedofX(float val)
	{	speedofX = val;	}
	void setIntensityofY(int val)
	{	intensityofY = val; }
	public void setStrokeWidth(float val)
	{	paint.setStrokeWidth(val); }
	float getX()
	{	return _x; }
	float getY()
	{	return _y; }
	float getSpeedofX()
	{	return speedofX; }
	int getIntensityofY()
	{	return intensityofY; } 
}